from .nth_variation import nth_variation
def minkowsky(a:dict, b:dict, p=2)->float:
    if p == 0:
        raise ValueError("Parameter p must be non zero.")
    """Returns the euclideam distance beetween a and b"""
    return round((nth_variation(a,b,p)**(1/p)),14)


def even_nth(a_val:float, b_val:float)->float:
    return a_val-b_val

def odd_nth(a_val:float, b_val:float)->float:
    return abs(a_val-b_val)

def nth_variation(a:dict, b:dict, exponent=2, overlap = False)->float:
    """Returns the nth power distance beetween a and b"""
    total = 0
    n = 0
    bget = b.get

    if exponent%2==0:
        nth = even_nth
    else:
        nth = odd_nth

    for k, a_val in a.items():
        b_val = bget(k)
        if b_val:
            total += nth(a_val, b_val)**exponent
        else:
            total += a_val**exponent
        n+=1
    for k, b_val in b.items():
        if k not in a:
            total += b_val**exponent
            n+=1
    result = round(total,14)
    if overlap:
        return result, n
    return result


from .bhattacharyya import bhattacharyya
from .canberra import canberra
from .chebyshev import chebyshev
from .cosine import cosine
from .euclidean import euclidean
from .hamming import hamming
from .hellinger import hellinger
from .jensen_shannon import jensen_shannon
from .kullback_leibler import kullback_leibler
from .mae import mae
from .manhattan import manhattan
from .minkowsky import minkowsky
from .mse import mse
from .nth_variation import nth_variation
from .pearson import pearson
from .squared_variation import squared_variation
from .total_variation import total_variation



from .distances_utils import sort
from math import sqrt
def hellinger(a:dict, b:dict) -> float:
    """Determines the Hellinger distance"""
    total = 0
    big, small = sort(a,b)
    big_get = big.get
    for key, small_value in small.items():
        big_value = big_get(key)
        if big_value:
            total += (sqrt(small_value)-sqrt(big_value))**2
    return round(sqrt(total)/sqrt(2),14)


from .distances_utils import sort
from math import log

def jensen_shannon(a:dict, b:dict)->float:
    """Returns the jensen shannon divergence beetween a and b"""
    total = 0
    delta = 0
    big, small = sort(a,b)

    big_get = big.get

    for key, value in small.items():
        ov = big_get(key)
        if ov:
            denominator = (ov + value)/2
            total += value*log(value/denominator) + ov*log(ov/denominator)
            delta -= ov
        else:
            delta += value

    total += (1+delta)*log(2)
    return round(total/2,14)


def canberra(a:dict, b:dict)->float:
    """Returns the canberra distance beetween a and b"""
    total = 0
    aget = a.get
    bget = b.get
    for k in a:
        aval = aget(k)
        bval = bget(k)
        if bval:
            total += abs(aval-bval)/(aval+bval)
        else:
            total += 1
    for k in b:
        if k not in a:
            total += 1
    return round(total,14)


from .distances_utils import sort
from math import log
def kullback_leibler(a:dict, b:dict) -> float:
    """Determines the Kullback–Leibler divergence"""
    total = 0
    big, small = sort(a,b)
    big_get = big.get
    for key, small_value in a.items():
        big_value = big_get(key)
        if big_value:
            total += small_value*log(small_value/big_value)
    return round(total,14)


def hamming(a:dict, b:dict)->float:
    """Returns the hamming distance beetween a and b"""
    common = 0
    for k in a:
        if k in b:
            common += 1
    return len(a) + len(b) - common*2


from math import sqrt
def cosine(a:dict, b:dict)->float:
    """Returns the cosine distance beetween a and b"""
    prod_ab = 0
    sqr_a = 0
    sqr_b = 0
    bget = b.get
    for k, a_val in a.items():
        b_val = bget(k,0)
        sqr_a += a_val**2
        if b_val:
            prod_ab += b_val*a_val

    for k, b_val in b.items():
        sqr_b += b_val**2

    return 1-round(prod_ab/(sqrt(sqr_a)*sqrt(sqr_b)),14)



from .distances_utils import sort
from math import sqrt, log

def bhattacharyya_coefficient(a:dict, b:dict) -> float:
    """Determines the bhattacharyya coefficient"""
    total = 0
    big, small = sort(a,b)
    big_get = big.get
    for k,small_value in small.items():
        big_value = big_get(k)
        if big_value:
            total += sqrt(big_value*small_value)
    return total

def bhattacharyya(a:dict, b:dict) -> float:
    """Determines the bhattacharyya distance"""
    return round(-log(bhattacharyya_coefficient(a,b)),14)


def sort(a:dict,b:dict):
    """Returns the bigger dict, then the smaller one"""
    if len(a)>len(b):
        return a,b
    return b,a

def mean(a,b,distance):
    result, n = distance(a,b,overlap=True)
    return round(result/n,14)


from math import sqrt
def pearson(a:dict, b:dict)->float:
    """Returns the nth power distance beetween a and b"""
    bget = b.get

    ab = 0
    n_mul = 0
    a_sum = 0
    b_sum = 0
    a_sum2 = 0
    b_sum2 = 0

    for k, a_val in a.items():
        b_val = bget(k)
        a_sum += a_val
        a_sum2 += a_val**2
        if b_val:
            n_mul += 1
            ab += a_val*b_val

    for k, b_val in b.items():
        b_sum += b_val
        b_sum2 += b_val**2

    len_a = len(a)
    len_b = len(b)

    a_mean = a_sum/len_a
    b_mean = b_sum/len_b

    return 1-round((ab - n_mul*a_mean*b_mean) / (sqrt(a_sum2 - len_a*a_mean**2)*sqrt(b_sum2 - len_b*b_mean**2)),14)


def chebyshev(a:dict, b:dict) -> float:
    """Determines the Chebyshev distance"""
    result = 0
    bget = b.get
    for key, a_val in a.items():
        b_val = bget(key)
        if b_val:
            result = max(result, abs(a_val - b_val))
        else:
            result = max(result, a_val)
    for key, b_val in b.items():
        if key not in a:
            result = max(result, b_val)
    return round(result,14)